#!C:\ci\holoviews_1522360513931\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'holoviews==1.9.5','console_scripts','holoviews'
__requires__ = 'holoviews==1.9.5'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('holoviews==1.9.5', 'console_scripts', 'holoviews')()
    )
